#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#  Copyright (c) 2019.       Mike Herbert
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA
"""
Provide functions to normalize text strings by converting to lowercase, removing noisewords.   
This is used by the lookup functions and the database build functions and match scoring  
 
noise_words is a list of replacements only used for match scoring   
phrase_cleanup is a list of replacements for db build, lookup and match scoring   
"""
import functools
import math
import sys
from re import sub

from geodata import GeodataBuild, Loc, GeoUtil

# Todo -  make all of these list driven

CACHE_SIZE = 30000


stop_words = {
    # list of stop words
    # These are forced to sort at end of Soundex name so that search for 'north pinewood' can be done as 'pinewood%' and
    # will find 'pinewood north'
    'east',
    'west',
    'north',
    'south',
    'the',
    'of',
    'de',
    'du',
    'town',
    'city',
    'county',
    'and',
    'on',
    'by',
    'new',
    'old',
    'big',
    'little',
    'royal',
    'borough',
    'department',
    'regional',
    'region',
    'district',
    'central,',
    'village',
    }

noise_words = [
    # apply this list of regex substitutions for match scoring
    # (r'), '), '),'),
    (r'normandy american '                     , 'normandie american '),
    (r'nouveau brunswick'                      , ' '),
    (r'westphalia'                             , 'westfalen'),
    (r'departement'                            , 'department'),
    (r'royal borough of windsor and maidenhead', 'berkshire'),
    (r'regional municipality'                  , 'county'),
    (r'kathedrale'                             , 'cathedral'),
    (r'citta metropolitana di '                , ' '),
    (r'kommune'                                , ''),
    (r"politischer bezirk "                    , ' '),
    (r'regional'                               , ' '),
    (r'region'                                 , ' '),
    (r'abbey'                                  , 'abbey'),
    (r'priory'                                 , 'abbey'),
    (r'greater'                                , ' '),
    (r' de '                                   , ' '),
    (r' di '                                   , ' '),
    (r' du '                                   , ' '),
    (r' of '                                   , ' '),

    (r"l'"                                     , ''),

    (r'erry'                                   , 'ury'),
    (r'ery'                                    , 'ury'),
    (r'borg'                                   , 'burg'),
    (r'bourg'                                  , 'burg'),
    (r'urgh'                                   , 'urg'),
    (r'mound'                                  , 'mund'),
    (r'ourne'                                  , 'orn'),
    (r'ney'                                    , 'ny'),
    ]

phrase_cleanup = [
    # always apply this list of regex substitutions 
    (r'  +'                                             , ' '),  # Strip multiple space to single space
    (r'\bmt '                                           , 'mount '),

    (r'\br\.k\. |\br k '                                , 'roman catholic '),
    (r'\brooms katholieke\b'                            , 'roman catholic'),

    (r'sveti |saints |sainte |sint |saint |sankt |st\. ', 'st '),  # Normalize Saint to St
    (r' co\.'                                           , ' county'),  # Normalize County
    (r'united states of america'                        , 'usa'),  # Normalize to USA   begraafplaats
    (r'united states'                                   , 'usa'),  # Normalize to USA

    (r'cimetiere'                                       , 'cemetery'),  # 
    (r'begraafplaats'                                   , 'cemetery'),  # 

    (r'town of '                                        , ' '),  # - remove town of
    (r'city of '                                        , ' '),  # - remove city of

    (r'village of ([^,]+)', r'\g<1> village'),  # Normalize 'Village of X' to 'X Village'
    (r'county of ([^,]+)'                               , r'\g<1> county'),  # Normalize 'Township of X' to 'X Township'
    (r'township of ([^,]+)'                             , r'\g<1> township'),  # Normalize 'Township of X' to 'X Township'
    (r'cathedral of ([^,]+)'                            , r'\g<1> cathedral'),  # Normalize 'Township of X' to 'X Township'
    (r'palace of ([^,]+)'                               , r'\g<1> palace'),  # Normalize 'Township of X' to 'X Township'
    (r'castle of ([^,]+)'                               , r'\g<1> castle'),  # Normalize 'Township of X' to 'X Township'

    (r"'(\w{2,})'"                                      , r"\g<1>"),  # remove single quotes around word, but leave apostrophes
    ]


no_punc_remove_commas = [
    # Regex to remove most punctuation including commas
    (r"[^a-z0-9 $*']+", " ")
    ]

no_punc_keep_commas = [
    # Regex to remove most punctuation but keep commas
    (r"[^a-z0-9 $*,']+" , " ")
    ]

local_country_names = {
    # Dictionary of key=local name, val = english name for country
    'norge'       : 'norway',
    'sverige'     : 'sweden',
    'osterreich'  : 'austria',
    'belgie'      : 'belgium',
    'brasil'      : 'brazil',
    'danmark'     : 'denmark',
    'eire'        : 'ireland',
    'magyarorszag': 'hungary',
    'italia'      : 'italy',
    'espana'      : 'spain',
    'deutschland' : 'germany',
    'prussia'     : 'germany',
    'suisse'      : 'switzerland',
    'schweiz'     : 'switzerland',
    }

ALIAS_FEAT = 2
ALIAS_ISO = 1
ALIAS_NAME = 0

alias_list = {
    # Dictionary of key=local name, val = (english name, iso, feature) for country, province, county
    'norge'               : ('norway', '', 'ADM0'),
    'sverige'             : ('sweden', '', 'ADM0'),
    'osterreich'          : ('austria', '', 'ADM0'),
    'belgie'              : ('belgium', '', 'ADM0'),
    'brasil'              : ('brazil', '', 'ADM0'),
    'danmark'             : ('denmark', '', 'ADM0'),
    'eire'                : ('ireland', '', 'ADM0'),
    'magyarorszag'        : ('hungary', '', 'ADM0'),
    'italia'              : ('italy', '', 'ADM0'),
    'espana'              : ('spain', '', 'ADM0'),
    'deutschland'         : ('germany', '', 'ADM0'),
    'prussia'             : ('germany', '', 'ADM0'),
    'suisse'              : ('switzerland', '', 'ADM0'),
    'schweiz'             : ('switzerland', '', 'ADM0'),

    'bayern'              : ('bavaria', 'de', 'ADM1'),
    'westphalia'          : ('westfalen', 'de', 'ADM1'),

    'normandy'            : ('normandie', 'fr', 'ADM1'),
    'brittany'            : ('bretagne', 'fr', 'ADM1'),
    'burgundy'            : ('bourgogne franche comte', 'fr', 'ADM1'),
    'franche comte'       : ('bourgogne franche comte', 'fr', 'ADM1'),
    'aquitaine'           : ('nouvelle aquitaine', 'fr', 'ADM1'),
    'limousin'            : ('nouvelle aquitaine', 'fr', 'ADM1'),
    'poitou charentes'    : ('nouvelle aquitaine', 'fr', 'ADM1'),
    'alsace'              : ('grand est', 'fr', 'ADM1'),
    'champagne ardenne'   : ('grand est', 'fr', 'ADM1'),
    'lorraine'            : ('grand est', 'fr', 'ADM1'),
    'languedoc roussillon': ('occitanie', 'fr', 'ADM1'),
    'nord pas de calais'  : ('hauts de france', 'fr', 'ADM1'),
    'picardy'             : ('hauts de france', 'fr', 'ADM1'),
    'auvergne'            : ('auvergne rhone alpes', 'fr', 'ADM1'),
    'rhone alpes'         : ('auvergne rhone alpes', 'fr', 'ADM1'),

    'breconshire'         : ('sir powys', 'gb', 'ADM2'),
    }


class Normalize:
    def __init__(self):
        # Build compiled lists of regex statements that will be used for normalization
        
        # phrase_rgx_remove_commas - Combine phrase dictionary and no punctuation_remove_commas and compile regex
        self.phrase_rgx_remove_commas = GeoUtil.MultiRegex(no_punc_remove_commas + phrase_cleanup)

        # phrase_rgx_keep_commas - Combine phrase dictionary and no punctuation_keep_commas and compile regex
        self.phrase_rgx_keep_commas = GeoUtil.MultiRegex(no_punc_keep_commas + phrase_cleanup)

        # noise_rgx  - Combine phrase dictionary with Noise words dictionary and compile regex (this is used for match scoring)
        self.noise_rgx = GeoUtil.MultiRegex(no_punc_keep_commas + phrase_cleanup + noise_words)
    
    @functools.lru_cache(maxsize=CACHE_SIZE)
    def normalize(self,text: str, remove_commas: bool) -> str:
        """
        Normalize text - Convert to lowercase ascii, remove most punctuation, apply replacements in phrase_cleanup list
        Remove commas if parameter set.   
        Remove all non alphanumeric except $ and *  
        Then call _phrase_normalize() which normalizes common phrases with multiple spellings, such as saint to st   
        #Args:   
            text:  Text to normalize   
            remove_commas:   True if commas should be removed   
    
        #Returns:   
            Normalized text
    
        """
    
        # remove all non alphanumeric except $ and * and comma(if flag set)
        if remove_commas:
            text = self.phrase_rgx_remove_commas.sub(text)
        else:
            text = self.phrase_rgx_keep_commas.sub(text)
        
        return text.strip()
    
    
    @functools.lru_cache(maxsize=CACHE_SIZE)
    def normalize_for_scoring(self, text: str) -> str:
        """
            Normalize the text for closeness scoring.  Apply normal normalization and then replacements for scoring_noise_words
    
        #Args:
            text: text to normalize
        #Returns:
    
        """
        #text = self.normalize(text=text, remove_commas=False)
        text = self.noise_rgx.sub(text)
        return text
    
    
    def _phrase_normalizeZZZ(self,text: str) -> str:
        """ Strip spaces and normalize spelling for items such as Saint and County """
        # Replacement patterns to clean up entries
        for pattern, replace in phrase_cleanup:
            text = sub(pattern, replace, text)
    
        return text
    
    
    def admin1_normalize(self,admin1_name: str, iso):
        """ Normalize historic or colloquial Admin1 names to current geoname standard """
        admin1_name = self.normalize(admin1_name, False)
        if iso == 'de':
            admin1_name = sub(r'bayern', 'bavaria', admin1_name)
            admin1_name = sub(r'westphalia', 'westfalen', admin1_name)
        elif iso == 'fr':
            admin1_name = sub(r'normandy', 'normandie', admin1_name)
            admin1_name = sub(r'brittany', 'bretagne', admin1_name)
            admin1_name = sub(r'burgundy', 'bourgogne franche comte', admin1_name)
            admin1_name = sub(r'franche comte', 'bourgogne franche comte', admin1_name)
            admin1_name = sub(r'aquitaine', 'nouvelle aquitaine', admin1_name)
            admin1_name = sub(r'limousin', 'nouvelle aquitaine', admin1_name)
            admin1_name = sub(r'poitou charentes', 'nouvelle aquitaine', admin1_name)
            admin1_name = sub(r'alsace', 'grand est', admin1_name)
            admin1_name = sub(r'champagne ardenne', 'grand est', admin1_name)
            admin1_name = sub(r'lorraine', 'grand est', admin1_name)
            admin1_name = sub(r'languedoc roussillon', 'occitanie', admin1_name)
            admin1_name = sub(r'midi pyrenees', 'occitanie', admin1_name)
            admin1_name = sub(r'nord pas de calais', 'hauts de france', admin1_name)
            admin1_name = sub(r'picardy', 'hauts de france', admin1_name)
            admin1_name = sub(r'auvergne', 'auvergne rhone alpes', admin1_name)
            admin1_name = sub(r'rhone alpes', 'auvergne rhone alpes', admin1_name)
        elif iso == 'ca':
            admin1_name = sub(r'brunswick', 'brunswick*', admin1_name)
    
        return admin1_name
    
    
    def admin2_normalize(self,admin2_name: str, iso) -> (str, bool):
        """
            Normalize historic or colloquial Admin2 names to standard
    
        Args:
            admin2_name: 
            iso: 
    
        Returns: TUPLE (result, modified) - result is new string, modified - True if modified
    
        """
        admin2_name = self.normalize(admin2_name, False)
    
        mod = False
    
        if iso == 'gb':
            admin2_name = sub(r'breconshire', 'sir powys', admin2_name)
            mod = True
    
        return admin2_name, mod
    
    
    def country_normalize(self,country_name) -> (str, bool):
        """
        normalize local language Country name to standardized English country name for lookups
        :param country_name:
        :return: (result, modified)
        result - new string
        modified - True if modified
        """
        country_name = sub(r'\.', '', country_name)  # remove .
    
        if local_country_names.get(country_name):
            country_name = local_country_names.get(country_name)
            return country_name.strip(' '), True
        else:
            return country_name.strip(' '), False
        
        
    def feature_by_population(self,pop, feat):
        # list of feature names indexed by log10 of population.  Small locations keep the original feature name
        feat_names = ['PPL', 'PPL', 'PPL', 'PPL',  'P10K', 'P1HK', 'PP1M', 'PP1M', 'PP1M']
        if  'PP' == feat[0:2] and pop>0:
            return(feat_names[int(math.log10(pop))])
        else:
            return(feat)
    
    def feature_normalize(self,feature:str, name:str, pop: int):
        """
        Normalize feature codes: 1) cleanup features for RUIN, HSTS, AREA,  
        2) set city feature based on population,
        3) change church, synagogue, etc to RLG 
        Args:
            feature: geoname feature code
            name: name of location
            pop: population of location
    
        Returns:
            normalized feature code
    
        """
        if feature[0:2] == 'PP':
            # This is a populated place.  Create new Feature code based on city population 
            feature = self.feature_by_population(pop, feature)
        elif feature == 'AREA' and 'island' in name:
                feature = 'ISL'
        elif feature == 'HSTS' or feature == 'RUIN':
            # Set feature type for Abbey/Priory, Castle, and Church if feature is RUIN or HSTS and name matches
            if 'abbey' in name :
                feature = 'MSTY' 
            elif 'priory' in name:
                feature = 'MSTY' 
            elif 'castle' in name:
                feature = 'CSTL'
            elif 'church' in name:
                feature = 'CH'
                
        if feature in ['CH', 'SYG', 'MSTY', 'MSQE', 'CTRR','CVNT', ]:
            feature = 'RLG'
            
        return feature
    
    """ 
    @functools.lru_cache(maxsize=CACHE_SIZE)
    def _remove_noise_wordsZZZ(self,text: str):
        # Calculate score with noise words removed    
        for pattern, replace in scoring_noise_words:
            text = sub(pattern, replace, text)
        return text
    """
    
    
    def remove_aliase(self,input_words, res_words) -> (str, str):
        if "middlesex" in input_words and "greater london" in res_words:
            input_words = sub('middlesex', 'greater london', input_words)
        return input_words, res_words
    
    
    def add_aliases_to_db(self,geo_build: GeodataBuild):
        #  Add alias names to DB
        for ky in alias_list:
            self.add_alias_to_db(ky, geo_build)
    
    
    def add_alias_to_db(self,ky: str, geo_build: GeodataBuild):
        alias_row = alias_list.get(ky)
        place = Loc.Loc()
        place.country_iso = alias_row[ALIAS_ISO].lower()
        place.city = alias_row[ALIAS_NAME]
        place.feature = alias_row[ALIAS_FEAT]
        place.place_type = Loc.PlaceType.CITY
    
        # Lookup main entry and get GEOID
        geo_build.geodb.s.lookup_place(place)
        if len(place.georow_list) > 0:
            if len(place.georow_list[0]) > 0:
                geo_row = list(place.georow_list[0][0:GeoUtil.Entry.SDX + 1])
                geo_build.update_geo_row_name(geo_row=geo_row, name=ky)
                geo_tuple = tuple(geo_row)
                geo_build.insert(geo_tuple=geo_tuple, feat_code=alias_row[ALIAS_FEAT])

@functools.lru_cache(maxsize=CACHE_SIZE)
def sorted_normalize(text):
    # Remove l' and d'  
    text = sub(r"l\'", "", text)
    text = sub(r"d\'", "", text)

    # Modify phrase X of Y to be Y X.  County of Whitley becomes Whitley County
    text = sub('([^,]+) of ([^,]+)', r'\g<2> \g<1>', text)

    # Force stop words to sort at end of list so wildcard will find phrases without them  (East, West, North, South)
    # North Haden becomes Haden North and Haden% will match it.
    word_list = sorted(text.split(' '), key=stop_words_last)
    return word_list

def stop_words_last(item):
    """
    key function for sorting - return key but force stop words to sort last
    Stop words come from Normalize.stop_words
    Args:
        item: 

    Returns:
        key for item, with stop words forced last
    """
    if item in stop_words:
        # Force stop words to sort last.  { will sort last
        return '{' + item
    return item

def deb(msg=None):
    # Display debug message with line number
    print(f"Debug {sys._getframe().f_back.f_lineno}: {msg if msg is not None else ''}")




#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#  Copyright (c) 2019.       Mike Herbert
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA

""" Build SQL where clause and arg list """
import re


class QueryItem:
    def __init__(self):
        self.where = ''
        self.args = tuple()
        self.table = 'main.geodata'

    def clear(self):
        self.where = ''
        self.args = tuple()

    def add_clauses(self, where_clauses: [str], terms: [], table=''):
        """
        Create a Where clause (and args list) by appending each item in clause with AND as separator
        Use equal, unless there is an '*', then use LIKE type clause
        Args:
            where_clauses: list of column names
            terms: list of values
            table to use
        Returns: None
        """
        if table != '':
            self.table = table

        for idx, where_clause in enumerate(where_clauses):
            # Walk thru the where clauses specified and add each to our query and list of args
            term = terms[idx]
            # If where clause is for feature column and term is ADM0 or ADM1, then we need to look into admin table
            if where_clause == 'feature' and table == '':
                if 'ADM0' in term or 'ADM1' in term:
                    self.table = 'main.admin'

            if len(term) > 0:
                if len(self.where) > 0:
                    # There is already text in Where, so add AND
                    self.where += ' AND '

                if term[-1] == '*':
                    # TODO handle case with multiple wildcards with one at end
                    # Wildcard search at end of text - Use >= and < (usually gives better performanc than LIKE)
                    term = self.create_wildcard(term, remove=True)
                    self.where += f' ({where_clause} >= ? and {where_clause} < ?) '
                    self.args += (term, inc_key(term),)
                elif '*' in term:
                    # Wildcard search in middle or start of search - Use LIKE
                    term = self.create_wildcard(term, remove=False)
                    self.where += f' ({where_clause} like ?) '
                    self.args += (term,)
                else:
                    self.where += f' {where_clause} = ?'
                    self.args += (term,)

    @staticmethod
    def create_wildcard(pattern: str, remove: bool):
        """
        Convert lookup pattern into a SQL lookup (convert * to %)
        Args:
            pattern: 
            remove: If True, remove  wildcards

        Returns:

        """
        
        # Create SQL wildcard pattern (convert * to %).  Add % on end
        if remove:
            return re.sub(r"\*", "", pattern)
        else:
            return re.sub(r"\*", "%", pattern)


def inc_key(text):
    """ increment the last letter of text by one.  Used to replace key in SQL LIKE case with less than """
    return text[0:-1] + chr(ord(text[-1]) + 1)



from geodata import GeoUtil, Normalize

places = [
    '12 baker st, Man!@#%^&(chester, , England',  # punctuation
    'department kathedrale of westphalia kommune ',  # normandie
    'archipel de saint augustin',
    'Le Mont Saint Michel',  # county of
    ]


noise_words = [
    # apply this list of regex substitutions for match scoring
    # (r'), '), '),'),
    (r'normandy american '                     , 'normandie american '),
    (r'nouveau brunswick'                      , ' '),
    (r'westphalia'                             , 'westfalen'),
    (r'departement'                            , 'department'),
    (r'royal borough of windsor and maidenhead', 'berkshire'),
    (r'regional municipality'                  , 'county'),
    (r'kathedrale'                             , 'cathedral'),
    (r'citta metropolitana di '                , ' '),
    (r'kommune'                                , ''),
    (r"politischer bezirk "                    , ' '),
    (r'regional'                               , ' '),
    (r'region'                                 , ' '),
    (r'abbey'                                  , 'abbey'),
    (r'priory'                                 , 'abbey'),
    (r'greater'                                , ' '),
    (r' de '                                   , ' '),
    (r' di '                                   , ' '),
    (r' du '                                   , ' '),
    (r' of '                                   , ' '),

    (r"l'"                                     , ''),

    (r'erry'                                   , 'ury'),
    (r'ery'                                    , 'ury'),
    (r'borg'                                   , 'burg'),
    (r'bourg'                                  , 'burg'),
    (r'urgh'                                   , 'urg'),
    (r'mound'                                  , 'mund'),
    (r'ourne'                                  , 'orn'),
    (r'ney'                                    , 'ny'),
    ]

phrase_cleanup = [
    # always apply this list of regex substitutions 
    (r'  +'                                             , ' '),  # Strip multiple space to single space
    (r'\bmt '                                           , 'mount '),

    (r'\br\.k\. |\br k '                                , 'roman catholic '),
    (r'\brooms katholieke\b'                            , 'roman catholic'),

    (r'sveti |saints |sainte |sint |saint |sankt |st\. ', 'st '),  # Normalize Saint to St
    (r' co\.'                                           , ' county'),  # Normalize County
    (r'united states of america'                        , 'usa'),  # Normalize to USA   begraafplaats
    (r'united states'                                   , 'usa'),  # Normalize to USA

    (r'cimetiere'                                       , 'cemetery'),  # 
    (r'begraafplaats'                                   , 'cemetery'),  # 

    (r'town of '                                        , ' '),  # - remove town of
    (r'city of '                                        , ' '),  # - remove city of

    (r'county of ([^,]+)'                               , r'\g<1> county'),  # Normalize 'Township of X' to 'X Township'
    (r'township of ([^,]+)'                             , r'\g<1> township'),  # Normalize 'Township of X' to 'X Township'
    (r'cathedral of ([^,]+)'                            , r'\g<1> cathedral'),  # Normalize 'Township of X' to 'X Township'
    (r'palace of ([^,]+)'                               , r'\g<1> palace'),  # Normalize 'Township of X' to 'X Township'
    (r'castle of ([^,]+)'                               , r'\g<1> castle'),  # Normalize 'Township of X' to 'X Township'

    (r"'(\w{2,})'"                                      , r"\g<1>"),  # remove single quotes around word, but leave apostrophes
    ]


no_punc_remove_commas = [
    # Regex to remove most punctuation including commas
    (r"[^a-z0-9 $*']+", " ")
    ]

no_punc_keep_commas = [
    # Regex to remove most punctuation but keep commas
    (r"[^a-z0-9 $*,']+" , " ")
    ]

"""
    r"[^a-z0-9 $*,']+" , " "

"""

# Regex to remove most punctuation including commas


# noise_rgx  - Combine phrase dictionary with Noise words dictionary and compile regex (this is used for match scoring)
# keys = sorted(dct.keys(), key=len, reverse=True)

phrase_rgx_keep_commas = GeoUtil.MultiRegex(no_punc_keep_commas + phrase_cleanup)
phrase_rgx_remove_commas = GeoUtil.MultiRegex(no_punc_remove_commas + phrase_cleanup)


#noise_rgx = GeoUtil.MultiRegex(phrase_cleanup + noise_words)
norm = Normalize.Normalize()

for txt in places:
    print(f'==== {txt} ====')
    print(f'  RESULT {phrase_rgx_remove_commas.sub(txt, lower=True, set_ascii=True)}')
    print(f' RESULT2 {norm.normalize(txt, False)}')
    



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#  Copyright (c) 2019.       Mike Herbert
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA
"""
geoname database support routines.  Add locations to geoname DB, create geoname tables and indices.   
Provides a number of methods to lookup locations by name, feature, admin ID, etc.
"""
import functools
import logging
import re

import phonetics

from geodata import Loc, Country, MatchScore, Normalize, QueryList
from geodata.GeoUtil import Query, Result, Entry, get_feature_group

FUZZY_LOOKUP = [Result.WILDCARD_MATCH, Result.WORD_MATCH, Result.SOUNDEX_MATCH]
CACHE_SIZE = 32768
COUNTRY_CACHE = 2048


class GeoSearch:
    """
    geoname database search routines.  look up items in geoname DB
    """

    def __init__(self, geodb):
        self.logger = logging.getLogger(__name__)
        self.detailed_debug = True
        self.start = 0
        self.use_wildcards = True
        self.total_lookups = 0
        self.cache = {}
        self.place_type = ''
        self.select_str = 'name, country, admin1_id, admin2_id, lat, lon, feature, geoid, sdx'
        self.geodb = geodb
        self.match = MatchScore.MatchScore()
        self.norm = Normalize.Normalize()
        self.place = Loc.Loc()

    def lookup_place(self, place: Loc) -> []:
        """
            **Lookup a place in geoname.org db**     
            Lookup is based on place.place_type as follows:  
                Loc.PlaceType.COUNTRY: does self.search_country(place)  
                Loc.PlaceType.ADVANCED_SEARCH: does self.feature_search(place)  
                Otherwise: do self.search_city(place)  
        # Args:   
            place: Loc instance.  Call Loc.parse_place() before calling lookup_place()   

        # Returns:   
            Best score found  
            place.georow_list contains a list of matching entries.  
            Each entry has: Lat, Long, districtID (County or State or Province ID), and a match quality score  

        """
        place.result_type = Result.STRONG_MATCH
        best_score = MatchScore.Score.VERY_POOR

        if place.place_type == Loc.PlaceType.COUNTRY:
            # Country
            if place.georow_list:
                place.country_name = self.get_country_name(place.country_name)
                best_score = MatchScore.Score.VERY_GOOD
        else:
            # General search 
            if place.place_type == Loc.PlaceType.ADMIN1:
                place.feature = "ADM1"
            place.georow_list.clear()
            best_score = self._search(place=place, georow_list=place.georow_list, name=place.city, admin1_id=place.admin1_id,
                                      admin2_id=place.admin2_id, iso=place.country_iso, feature=place.feature, sdx=get_soundex(place.city))
        self.logger.debug(f'**LOOKUP PLACE  score={best_score}')
        return best_score

    def deep_lookup(self, place: Loc) -> []:
        """
        Do a lookup based on soundex and combinations of words
        Args:
            place: 

        Returns: best score found

        """
        # Get more results unless we have a high score
        row_list = []
        best_score = self.search_for_combinations(row_list, place.city, place, 'main.geodata')
        if len(row_list) > 0:
            place.georow_list.extend(row_list)
        if best_score > MatchScore.Score.VERY_POOR - 10:
            # Search for each term in name
            row_list = []
            best_score = self.search_each_term(row_list, place.city, place, 'main.geodata')
            if len(row_list) > 0:
                place.georow_list.extend(row_list)
        return best_score

    def get_admin2_name(self, admin1_id, admin2_id, iso) -> str:
        """
        Search for Admin2 name using admin2_id and admin1_id 

        # Args:   
            admin1_id: Admin1 ID   
            admin2_id: Admin2 ID   
            iso: country ISO     

        # Returns:
            Admin2 name
        """
        return self._get_name(admin1_id=admin1_id, admin2_id=admin2_id, iso=iso, feature='ADM2')

    def get_admin1_name(self, admin1_id, iso) -> str:
        """
        Search for Admin1 name using admin1_id (rather than place instance)

        # Args:   
            admin1_id: Admin1 ID   
            iso: country ISO     

        # Returns:
            Admin1 name.   
        """
        return self._get_name(admin1_id=admin1_id, admin2_id='', iso=iso, feature='ADM1')

    def get_country_name(self, iso: str) -> str:
        """
             return country name for specified ISO code 
        # Args:   
            iso:   Country ISO code
        # Returns:
            Country name or ''
        """
        return self._get_name(admin1_id='', admin2_id='', iso=iso, feature='ADM0')

    @functools.lru_cache(maxsize=CACHE_SIZE)
    def _get_name(self, admin1_id, admin2_id, iso, feature, sdx='') -> str:
        """
             return  name for specified ID

        # Args:   
            iso:   Country ISO code

        # Returns:
             name or ''
        """
        row_list = []
        key = f'{admin1_id}_{admin2_id}_{iso}_{feature}'
        if len(admin1_id + admin2_id + iso) == 0:
            return ''

        self._search(georow_list=row_list, place=None, name='', admin1_id=admin1_id, admin2_id=admin2_id, iso=iso, feature=feature, sdx=sdx)

        if len(row_list) > 0:
            self.cache[key] = row_list[0][Entry.NAME]
            return row_list[0][Entry.NAME]
        else:
            return ''

    def get_admin1_alternate_name(self, admin1_id, place: Loc) -> (str, str):
        """
             Get Admin1 name from alternate name table 

        # Args:   
            place:   place instance.  place.admin1_id is used for lookup

        # Returns:
            (admin1 name, lang)  place.admin2_id is updated with best match 
        """
        # self.logger.debug('GET ADMIN1 ALT NAME')
        if len(admin1_id) == 0:
            return '', ''

        # Find GEOID for admin1_id and then get alternate name for GEOID
        row_list = []
        self._search(georow_list=row_list, place=None, name='', admin1_id=admin1_id, admin2_id='', iso=place.country_iso, feature='ADM1')
        if len(row_list) > 0:
            row = row_list[0]
            # Get alternate name for this GEOID
            admin1_name, lang = self.get_alternate_name(row[Entry.ID])
            return admin1_name, lang
        else:
            return '', ''

    def get_alternate_name(self, geoid) -> (str, str):
        """
        Retrieve alternate name for specified GEOID   

        #Args:    
            geoid: Geoid to get alternate names for   

        #Returns: 
            (name, lang)   

        """
        # todo - make rowlist a param
        query_list = [
            Query(where="geoid = ?",
                  args=(geoid,),
                  result=Result.STRONG_MATCH)]
        select = 'name, lang'
        row_list = []
        self.geodb.process_query_list(result_list=row_list, place=None, select_fields=select, from_tbl='main.altname', query_list=query_list)
        if len(row_list) > 0:
            return row_list[0][0], row_list[0][1]
        else:
            return '', ''

    @functools.lru_cache(maxsize=COUNTRY_CACHE)
    def get_iso_from_admin1_id(self, admin1_id, country_iso) -> str:
        """
                Search for country iso using admin1_name

                # Args:   
                    admin1_name:
                    place:   

                # Returns:
                    admin1_id
                """
        row_list = []
        self.logger.debug(f'GET ISO ID for adm1 id [{admin1_id}]')
        self._search(georow_list=row_list, place=None, name='', admin1_id=admin1_id, admin2_id='', iso=country_iso, feature='ADM1', sdx='')
        if len(row_list) > 0:
            country_iso = row_list[0][Entry.ISO]
        else:
            country_iso = ''
        return country_iso

    @functools.lru_cache(maxsize=CACHE_SIZE)
    def get_admin1_id(self, admin1_name, country_iso) -> str:
        """
        Search for Admin1 ID using admin1_name

        # Args:   
            admin1_name:
            place:   

        # Returns:
            admin1_id
        """
        row_list = []
        self.place.clear()
        self.place.admin1_name = admin1_name
        self.place.country_iso = country_iso
        admin1_name = self.norm.admin1_normalize(admin1_name, country_iso)
        self.logger.debug(f'GET ADMIN1 ID from [{admin1_name}]')
        self._search(georow_list=row_list, place=None, name=admin1_name, admin1_id='', admin2_id='', iso=country_iso, feature='ADM1', sdx='')

        if len(row_list) == 0:
            # Nothing found.  Try deeper search
            self.logger.debug(f'not found - search for combo [{admin1_name}]')
            self.search_for_combinations(row_list=row_list, target=admin1_name, place=self.place, table='main.admin')
        else:
            self.logger.debug(f'found {row_list}')
            pass

        if len(row_list) > 0:
            admin1_id = row_list[0][Entry.ADM1]
        else:
            admin1_id = ''
        return admin1_id

    @functools.lru_cache(maxsize=COUNTRY_CACHE)
    def get_country_iso(self, country_name) -> str:
        """
             return country ISO code for place.country_name   

        # Args:   
            place:   place instance.  looks up by place.country_name   

        # Returns:   
            Country ISO or ''.     
        """
        row_list = []
        self.place.clear()
        self.place.country_name = country_name

        country_name, modified = self.norm.country_normalize(country_name)
        if len(country_name) == 0:
            return ''
        sdx = get_soundex(country_name) + '*'

        best = self._search(georow_list=row_list, place=self.place, name=country_name, admin1_id='', admin2_id='', iso='', feature='ADM0', sdx='')

        if self.place.result_type == Result.STRONG_MATCH:
            iso = row_list[0][Entry.ISO]
            self.place.country_name = row_list[0][Entry.NAME]
        else:
            # Lookup by soundex
            best = self._search(georow_list=row_list, place=self.place, name='', admin1_id='', admin2_id='', iso='', feature='ADM0',
                                sdx=sdx)
            if self.place.result_type == Result.STRONG_MATCH:
                iso = row_list[0][Entry.ISO]
                self.place.country_name = row_list[0][Entry.NAME]
            else:
                iso = ''
        return iso

    def update_names(self, place):
        # Use ID fields to fill in  missing names for admin1, admin2, and country
        if place.admin1_id != '':
            if place.admin1_name == '':
                place.admin1_name = self.get_admin1_name(place.admin1_id, place.country_iso)
            if place.admin2_name == '':
                place.admin2_name = self.get_admin2_name(place.admin1_id, place.admin2_id, place.country_iso)
        place.country_name = str(self.get_country_name(place.country_iso))

    def _search(self, georow_list, place, name, admin1_id, admin2_id, iso, feature, sdx=''):
        """

        Args:
            georow_list: Returns list of matching georows
            name: name of location to lookup
            admin1_id: admin1_id of location (if available, otherwise '')
            admin2_id: 
            iso: country iso of location (if available,  otherwise '')
            feature: feature code of location (if available, otherwise '')

        Returns: result_type

        """
        query_list = []

        # Put together where clauses with all non-blank fields 
        if len(name + admin2_id + admin2_id + admin1_id + iso) == 0:
            return Result.NO_MATCH
        ql = QueryList.QueryItem()

        where_clauses = ["name", "country", "admin1_id", "admin2_id", "feature"]
        terms = [name, iso, admin1_id, admin2_id, feature]
        ql.add_clauses(where_clauses=where_clauses, terms=terms)  # add 
        query_list.append(Query(where=ql.where, args=ql.args, result=Result.PARTIAL_MATCH))

        ql.clear()
        if len(sdx) > 0:
            where_clauses = ["country", "feature", "sdx"]
            terms = [iso, feature, sdx]
            ql.add_clauses(where_clauses=where_clauses, terms=terms)  # add 
            query_list.append(Query(where=ql.where, args=ql.args, result=Result.PARTIAL_MATCH))

        # See if we can determine feature type from name and lookup by that
        self.add_feature_query(query_list, name, iso)

        best = self.geodb.process_query_list(result_list=georow_list, place=place, select_fields=self.select_str, from_tbl=ql.table,
                                             query_list=query_list, debug=True)
        return best

    def search_each_term(self, row_list, target, place, table):
        """
        Search for soundex of combinations of words in target
        Words are sorted and then Soundex searched with every combination of one word missing
        Args:
            row_list: 
            target: 
            place: 
            table: 

        Returns:

        """
        query_list = []
        best = 999

        sdx = get_soundex(target)
        if len(sdx) > 3:
            sdx_word_list = sdx.split(' ')
        else:
            sdx_word_list = ['']

        if len(sdx_word_list) > 1 and len(place.country_iso) > 0:
            #  Try each word 
            self.logger.debug(f'Search EACH {sdx_word_list}')
            for idx, word in enumerate(sdx_word_list):
                if len(word) == 0:
                    continue
                pattern = word[0:4]
                self.logger.debug(f' {pattern}')
                if place.feature:
                    search, pattern, pattern2 = convert_like(column_name='sdx', pattern=pattern)
                    where = f'{search} AND  country = ? AND feature = ?'
                    if pattern2:
                        args = (pattern, pattern2, place.country_iso, place.feature,)
                    else:
                        args = (pattern, place.country_iso, place.feature,)
                    query_list.append(Query(where=where, args=args, result=Result.SOUNDEX_MATCH))
                else:
                    search, pattern, pattern2 = convert_like(column_name='sdx', pattern=pattern)
                    where = f'{search} AND  country = ? '
                    if pattern2:
                        args = (pattern, pattern2, place.country_iso,)
                    else:
                        args = (pattern, place.country_iso,)
                    query_list.append(Query(where=where, args=args, result=Result.SOUNDEX_MATCH))

            best = self.geodb.process_query_list(result_list=row_list, place=place, select_fields=self.select_str,
                                                 from_tbl=table, query_list=query_list, debug=True)
        return best

    def search_for_combinations(self, row_list, target, place, table):
        """
        Search for soundex of combinations of words in target
        Words are sorted and then Soundex searched of every combination with one word missing
        Args:
            row_list: 
            target: 
            place: 
            table: 

        Returns: None

        """
        query_list = []
        best = 999

        sdx = get_soundex(target)

        if len(sdx) > 3 and len(place.country_iso) > 0:
            sdx_word_list = sdx.split(' ')
            self.logger.debug(f'COMBO SEARCH {sdx_word_list}')

            if len(sdx_word_list) == 1:
                if place.feature:
                    query_list.append(Query(where="(sdx >= ? and sdx < ?) AND country = ? AND feature = ?",
                                            args=(sdx, inc_key(sdx), place.country_iso, place.feature,),
                                            result=Result.SOUNDEX_MATCH))
                else:
                    query_list.append(Query(where="(sdx >= ? and sdx < ?) AND country = ?",
                                            args=(sdx, inc_key(sdx), place.country_iso,),
                                            result=Result.SOUNDEX_MATCH))
            else:
                # Multiple words - Try every combination with a single word removed
                for ignore_word_idx in range(0, len(sdx_word_list)):
                    pattern = ''
                    for idx, word in enumerate(sdx_word_list):
                        if idx != ignore_word_idx:
                            if len(pattern) == 0:
                                pattern += word
                            else:
                                pattern += ' ' + word
                    # pattern += '%'
                    self.logger.debug(f'{ignore_word_idx}) {pattern}')
                    if place.feature:
                        query_list.append(Query(where="(sdx >= ? and sdx < ?) AND country = ? AND feature = ?",
                                                args=(pattern, inc_key(pattern), place.country_iso, place.feature,),
                                                result=Result.SOUNDEX_MATCH))
                    else:
                        query_list.append(Query(where="(sdx >= ? and sdx < ?) AND country = ?",
                                                args=(pattern, inc_key(pattern), place.country_iso,),
                                                result=Result.SOUNDEX_MATCH))

            best = self.geodb.process_query_list(result_list=row_list, place=place, select_fields=self.select_str,
                                                 from_tbl=table, query_list=query_list, debug=True)

            self.logger.debug(f'search_for_combos ')
        return best

    def lookup_geoid(self, georow_list, geoid, place: Loc, admin=False) -> None:
        """
        Search by GEOID 
        Args:
            georow_list: 
            geoid: 
            place: 
            admin: If True, look up in admin table, otherwise look in geodata table

        Returns: None.  georow_list contains list of matches
        """
        if admin:
            table = 'main.admin'
        else:
            table = 'main.geodata'
        query_list = [
            Query(where="geoid = ? ",
                  args=(geoid,),
                  result=Result.STRONG_MATCH)
            ]
        self.geodb.process_query_list(result_list=georow_list, place=None, select_fields=self.select_str,
                                      from_tbl=table, query_list=query_list)

    def lookup_dbid(self, georow_list, dbid, place: Loc, admin=False) -> None:
        """
        Search by Database ID 
        Args:
            georow_list: 
            dbid: 
            place: 
            admin: If True, look up in admin table, otherwise look in geodata table

        Returns: None.  georow_list contains list of matches
        """
        if admin:
            table = 'main.admin'
        else:
            table = 'main.geodata'
        query_list = [
            Query(where="id = ? ",
                  args=(dbid,),
                  result=Result.STRONG_MATCH)
            ]
        self.geodb.process_query_list(result_list=georow_list, place=None, select_fields=self.select_str,
                                      from_tbl=table, query_list=query_list)
        if georow_list:
            # self.logger.debug(georow_list[0])
            self.copy_georow_to_place(georow_list[0], place, fast=True)
        else:
            self.logger.debug('no match')

    def copy_georow_to_place(self, row, place: Loc, fast: bool):
        """
        Copy data from DB row into place instance   
        Country, admin1_id, admin2_id, city, lat/lon, feature, geoid are updated if available   
        #Args:   
            row: georow from geoname database   
            place: Loc instance   
            fast: Currently ignored
        #Returns:   
            None.  Place instance is updated with data from georow   
        """
        place.admin1_id = ''
        place.admin2_id = ''
        place.admin1_name = ''
        place.admin2_name = ''
        place.city = ''

        place.country_iso = str(row[Entry.ISO])
        place.lat = row[Entry.LAT]
        place.lon = row[Entry.LON]
        place.feature = str(row[Entry.FEAT])
        # self.logger.debug(f'feat={place.feature}')
        place.geoid = str(row[Entry.ID])
        place.prefix = row[Entry.PREFIX]
        place.place_type = Loc.PlaceType.CITY

        if place.feature == 'ADM0':
            place.place_type = Loc.PlaceType.COUNTRY
            pass
        elif place.feature == 'ADM1':
            place.admin1_id = row[Entry.ADM1]
            place.place_type = Loc.PlaceType.ADMIN1
        elif place.feature == 'ADM2':
            place.admin1_id = row[Entry.ADM1]
            place.admin2_id = row[Entry.ADM2]
            place.place_type = Loc.PlaceType.ADMIN2
        else:
            place.admin1_id = row[Entry.ADM1]
            place.admin2_id = row[Entry.ADM2]
            place.city = row[Entry.NAME]

        if place.admin1_id != '':
            if place.admin1_name == '':
                place.admin1_name = self.get_admin1_name(place.admin1_id, place.country_iso)
            if place.admin2_name == '':
                place.admin2_name = self.get_admin2_name(place.admin1_id, place.admin2_id, place.country_iso)
        place.country_name = str(self.get_country_name(row[Entry.ISO]))

        if place.admin2_name is None:
            place.admin2_name = ''
        if place.admin1_name is None:
            place.admin1_name = ''

        place.city = str(place.city)
        if place.city is None:
            place.city = ''

        try:
            place.score = row[Entry.SCORE]
        except IndexError:
            pass

    @staticmethod
    def create_wildcard(pattern):
        """
        Create wildcard pattern.  Convert * to %.  Add % on end   
        #Args:   
            pattern:   

        #Returns: wildcard pattern   

        """
        if '*' in pattern:
            return re.sub(r"\*", "%", pattern)
        else:
            return f'%{pattern}%'

    @staticmethod
    def create_admin1_wildcard(pattern):
        """
        Create wildcard pattern.  Convert * to %.  Add % on end.  Currently this  
        is the same at create_wildcard   
        #Args:   
            pattern:    

        #Returns: wildcard pattern   

        """
        if '*' in pattern:
            return re.sub(r"\*", "%", pattern)
        else:
            return f'{pattern}%'

    @staticmethod
    def create_county_wildcard(pattern):
        """
        create wildcard for county lookup  
        #Args:   
            pattern:    

        #Returns:  

        """
        # Try pattern with 'shire' removed
        pattern = re.sub(r"shire", "", pattern)
        # Create SQL wildcard pattern (convert * to %).  Add % on end
        if '*' in pattern:
            return re.sub(r"\*", "%", pattern)
        else:
            return f'{pattern}'

    def set_display_names(self, temp_place):
        """
            See if there is an alternate name entry for this place   

        #Args:   
            temp_place: place instance   

        #Returns: None   

        """
        place_lang = Country.Country.get_lang(temp_place.country_iso)
        res, lang = self.get_alternate_name(temp_place.geoid)
        if res != '':  # and (lang == place_lang or lang == 'ut8'):
            temp_place.city = res

        res, lang = self.get_admin1_alternate_name(temp_place.city, temp_place)
        if res != '':  # and (lang == place_lang or lang == 'ut8'):
            temp_place.admin1_name = res

    def debugZZZ(self, text):
        if self.detailed_debug:
            self.logger.debug(text)

    def add_feature_query(self, query_list, target, iso):
        # todo use ql.add instead
        # Scan target to see if we can determine what feature type it is
        word, group = get_feature_group(target)
        if word != '':
            if 'st ' not in word:
                targ = re.sub(word, '', target).strip(' ')
            else:
                targ = target
            query_list.append(Query(where="(name >= ? and name < ?) AND country = ? AND feature = ?",
                                    args=(targ, inc_key(targ), iso, group),
                                    result=Result.WORD_MATCH))
            self.logger.debug(f'Add Feature query [{targ}] {group}')


def get_soundex(text) -> str:
    """
    Returns: Phonetics Double Metaphone Soundex code for sorted words in text  
    Words are alpha sorted but stop words are forced to end
    First two actual letters of word are prepended
    """
    sdx = []
    word_list = Normalize.sorted_normalize(text)

    for word in word_list:
        sdx.append(get_word_soundex(word))

    res = ' '.join(sdx)
    if len(res) == 0:
        res = text
    return res.lower()


@functools.lru_cache(maxsize=CACHE_SIZE)
def get_word_soundex(word):
    if len(word) > 1:
        return word[0:2] + phonetics.dmetaphone(word)[0]
    else:
        return phonetics.dmetaphone(word)[0]


def convert_like(pattern, column_name):
    """
    Convert SQL LIKE search to comparison if has % at end.  

    Args:
        pattern: 
        column_name: the column name for the search
    Returns: (txt, pattern, pattern2)
        1)  if % is not at end returns:
            txt is term like ?
            pattern1 is pattern
            pattern2 is ''
        or
        2) an equivalent SQL comparison if % at end:  
            txt is term >= ? and term < ?
            pattern1 is pattern
            pattern2 is term with last letter incremented
    """
    pattern2 = ''
    if pattern[-1] == '%':
        # Wildcard at end - Use  xx>= and yy< instead of LIKE to ensure Index can be used
        txt = f'({column_name} >= ? and {column_name} < ?)'
        pattern2 = inc_key(pattern)
    elif '%' in pattern:
        # Wildcard in middle - use LIKE
        txt = f'{column_name} LIKE ?'
    else:
        txt = f'{column_name} = ?'

    # print(f'[{txt}] ')
    return txt, pattern, pattern2


def inc_key(text):
    """ increment the last letter of text by one.  Used to replace key in SQL LIKE case with less than """
    if len(text) > 0:
        return text[0:-1] + chr(ord(text[-1]) + 1)
    else:
        return text
